from .fancy_indexing_list import FancyIndexingList
from .vec3 import Vec3, vec_type, mat_type
from .mat3 import Mat3
from .lattice_utils import (
    lat_params_to_lat_vecs,
    lat_vecs_to_lat_params,
    cart2frac,
    frac2cart,
)
from .symmetry_utils import symmop_from_xyz_str
from .html_utils import view_mol

__all__ = [
    "FancyIndexingList",
    "Vec3",
    "vec_type",
    "mat_type",
    "Mat3",
    "lat_params_to_lat_vecs",
    "lat_vecs_to_lat_params",
    "cart2frac",
    "frac2cart",
    "symmop_from_xyz_str",
    "view_mol",
]



import copy

from molgeom import Atom, Molecule, Vec3


def test_FancyIndexing_Molecule():
    a1 = Atom("N", 1.0, 1.0, 1.0)
    a2 = Atom("H", 1.0, 0.0, 0.0)
    a3 = Atom("H", 0.0, 1.0, 0.0)
    a4 = Atom("H", 0.0, 0.0, 1.0)
    mol = Molecule(a1, a2, a3, a4)

    assert mol[0] == a1
    assert mol[-1] == a4
    assert mol[[0, 1]] == Molecule(a1, a2)
    assert mol[(0, 1)] == Molecule(a1, a2)
    assert mol[0, 1] == Molecule(a1, a2)
    assert mol[0, -1] == Molecule(a1, a4)
    assert mol[[0, 1, 2]] == Molecule(a1, a2, a3)
    assert mol[[0, -1, 1, -2]] == Molecule(a1, a4, a2, a3)
    assert mol[1:] == Molecule(a2, a3, a4)
    assert mol[:3] == Molecule(a1, a2, a3)
    assert mol[-2:] == Molecule(a3, a4)
    assert mol[:-2] == Molecule(a1, a2)
    assert mol[1:3] == Molecule(a2, a3)
    assert mol[:] == Molecule(a1, a2, a3, a4)
    assert mol[1:3:2] == Molecule(a2)
    assert mol[1:4:2] == Molecule(a2, a4)
    assert mol[1::2] == Molecule(a2, a4)
    assert mol[:3:2] == Molecule(a1, a3)
    assert mol[::2] == Molecule(a1, a3)
    assert mol[::-1] == Molecule(a4, a3, a2, a1)
    mol[1] = Atom("C", 1.0, 1.0, 1.0)
    assert mol[1] == Atom("C", 1.0, 1.0, 1.0)
    mol[[1, 2]] = Molecule(Atom("C", 1.0, 1.0, 1.0), Atom("C", 1.0, 1.0, 1.0))
    assert mol[[1, 2]] == Molecule(Atom("C", 1.0, 1.0, 1.0), Atom("C", 1.0, 1.0, 1.0))
    mol[1, -1, 2] = Molecule(
        Atom("C", 1.0, 0.0, 0.0),
        Atom("C", 0.0, 1.0, 0.0),
        Atom("C", 0.0, 0.0, 1.0),
    )
    assert mol[1, -1, 2] == Molecule(
        Atom("C", 1.0, 0.0, 0.0),
        Atom("C", 0.0, 1.0, 0.0),
        Atom("C", 0.0, 0.0, 1.0),
    )
    assert mol[1:3] == Molecule(Atom("C", 1.0, 0.0, 0.0), Atom("C", 0.0, 0.0, 1.0))
    assert mol[0:4:2] == Molecule(Atom("N", 1.0, 1.0, 1.0), Atom("C", 0.0, 0.0, 1.0))
    mol = Molecule(a1, a2, a3, a4)
    mol[[0, 1, -1]] = Molecule(
        Atom("C", 1.0, 0.0, 0.0),
        Atom("C", 0.0, 1.0, 0.0),
        Atom("C", 0.0, 0.0, 1.0),
    )
    assert mol[[0, 1, -1]] == Molecule(
        Atom("C", 1.0, 0.0, 0.0),
        Atom("C", 0.0, 1.0, 0.0),
        Atom("C", 0.0, 0.0, 1.0),
    )
    assert mol == Molecule(
        Atom("C", 1.0, 0.0, 0.0),
        Atom("C", 0.0, 1.0, 0.0),
        Atom("H", 0.0, 1.0, 0.0),
        Atom("C", 0.0, 0.0, 1.0),
    )


def test_slice_id_check_Molecule():
    a1 = Atom("N", 1.0, 1.0, 1.0)
    a2 = Atom("H", 1.0, 0.0, 0.0)
    a3 = Atom("H", 0.0, 1.0, 0.0)
    a4 = Atom("H", 0.0, 0.0, 1.0)
    mol = Molecule(a1, a2, a3, a4)

    assert mol[:] == mol and mol[:] is not mol
    assert mol[1:3] == Molecule(a2, a3) and mol[1:3] is not Molecule(a2, a3)


def test_add_atom():
    mol1 = Molecule()
    a1 = Atom("C", 0, 1.1, 0.2)
    a2 = Atom("H", 0, 1.2, 0.3)
    a3 = Atom("H", 0, 1.3, 0.4)
    mol1.add_atom(a1)
    mol1.add_atoms_from([a2, a3])
    mol2 = Molecule(a1, a2, a3)

    assert len(mol1.atoms) == 3
    assert mol1.atoms[0] == a1
    assert mol1.atoms[1] == a2
    assert mol1.atoms[2] == a3
    assert mol1 == mol2


def test_copy():
    mol = Molecule()
    a1 = Atom("C", 0, 1.1, 0.2)
    a2 = Atom("H", 0, 1.2, 0.3)
    a3 = Atom("H", 0, 1.3, 0.4)
    mol.add_atoms_from([a1, a2, a3])
    mol_copied_from_method = mol.copy()
    mol_copied_from_copy = copy.deepcopy(mol)
    mol2 = mol

    assert mol == mol2
    assert mol is mol2
    assert mol == mol_copied_from_method
    assert mol is not mol_copied_from_method
    assert mol == mol_copied_from_copy
    assert mol is not mol_copied_from_copy


def test_is_same_geom():
    mol1 = Molecule()
    a1 = Atom("C", 0, 1.1, 0.2)
    a2 = Atom("H", 0, 1.2, 0.3)
    a3 = Atom("H", 0, 1.3, 0.4)
    mol1.add_atoms_from([a1, a2, a3])

    mol2 = Molecule()
    a2 = Atom("H", 0, 1.2, 0.3)
    a3 = Atom("H", 0, 1.3, 0.4)
    a1 = Atom("C", 0, 1.1, 0.2)
    mol2.add_atoms_from([a1, a2, a3])

    assert mol1.is_same_geom(mol2)

    mol2 = Molecule()
    a1 = Atom("C", 0, 1.1, 0.2)
    a2 = Atom("H", 0, 1.2, 0.3)
    a3 = Atom("H", 0, 1.3, 0.5)
    mol2.add_atoms_from([a1, a2, a3])

    assert not mol1.is_same_geom(mol2)


def test_replicate():
    a1 = Atom("O", 0, 0, 0)
    a2 = Atom("H", 1, 0, 0)
    a3 = Atom("H", 0, 1, 0)
    mol = Molecule(a1, a2, a3)
    mol.lattice_vecs = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    mol.replicate([0, 1], [0, 1], [0, 1])

    assert len(mol.atoms) == 3
    assert mol == Molecule(a1, a2, a3)

    mol = Molecule(a1, a2, a3)
    mol.lattice_vecs = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    mol_copied = mol.copy()
    mol_copied.translate(
        -1 * (mol.lattice_vecs[0] + mol.lattice_vecs[1] + mol.lattice_vecs[2])
    )
    mol.replicate([-1, 0], [-1, 0], [-1, 0])

    assert len(mol.atoms) == 3
    assert mol == mol_copied
    assert mol.lattice_vecs is not None

    mol.lattice_vecs = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    mol.replicate([-1, 2], [-1, 2], [-1, 2])
    assert len(mol.atoms) == 81


def test_is_inside_cell():
    a1 = Atom("N", 3, 3, 3)
    a2 = Atom("H", 1, 0, 0)
    a3 = Atom("H", 0, 1, 0)
    a4 = Atom("H", 0, 0, 1)
    mol = Molecule(a1, a2, a3, a4)
    mol.lattice_vecs = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    for atom in mol[1:]:
        assert mol.is_inside_cell(atom)
    assert not mol.is_inside_cell(mol[0])
    assert not mol.is_inside_cell(a1)

    mol = Molecule()
    mol.lattice_vecs = [Vec3(1.0, 0.0, 0.0), Vec3(0.5, 1.0, 0.0), Vec3(0.3, 0.2, 1.0)]
    inside_vec = (
        0.1 * mol.lattice_vecs[0]
        + 0.2 * mol.lattice_vecs[1]
        + 0.3 * mol.lattice_vecs[2]
    )
    outside_vec_1 = (
        1.1 * mol.lattice_vecs[0]
        + 0.2 * mol.lattice_vecs[1]
        + 0.3 * mol.lattice_vecs[2]
    )
    outside_vec_2 = (
        0.1 * mol.lattice_vecs[0]
        + 1.2 * mol.lattice_vecs[1]
        + 0.3 * mol.lattice_vecs[2]
    )
    outside_vec_3 = (
        0.1 * mol.lattice_vecs[0]
        + 0.2 * mol.lattice_vecs[1]
        + 1.3 * mol.lattice_vecs[2]
    )
    outside_vec_4 = (
        -0.1 * mol.lattice_vecs[0]
        + 0.2 * mol.lattice_vecs[1]
        + 0.3 * mol.lattice_vecs[2]
    )
    outside_vec_5 = (
        0.1 * mol.lattice_vecs[0]
        - 0.2 * mol.lattice_vecs[1]
        + 0.3 * mol.lattice_vecs[2]
    )
    outside_vec_6 = (
        0.1 * mol.lattice_vecs[0]
        + 0.2 * mol.lattice_vecs[1]
        - 0.3 * mol.lattice_vecs[2]
    )
    outside_vec_7 = (
        1.1 * mol.lattice_vecs[0]
        + 1.2 * mol.lattice_vecs[1]
        + 1.3 * mol.lattice_vecs[2]
    )
    outside_vec_8 = (
        -1.1 * mol.lattice_vecs[0]
        - 1.2 * mol.lattice_vecs[1]
        - 1.3 * mol.lattice_vecs[2]
    )
    assert mol.is_inside_cell(Atom.from_vec("H", inside_vec))
    assert not mol.is_inside_cell(Atom.from_vec("H", outside_vec_1))
    assert not mol.is_inside_cell(Atom.from_vec("H", outside_vec_2))
    assert not mol.is_inside_cell(Atom.from_vec("H", outside_vec_3))
    assert not mol.is_inside_cell(Atom.from_vec("H", outside_vec_4))
    assert not mol.is_inside_cell(Atom.from_vec("H", outside_vec_5))
    assert not mol.is_inside_cell(Atom.from_vec("H", outside_vec_6))
    assert not mol.is_inside_cell(Atom.from_vec("H", outside_vec_7))
    assert not mol.is_inside_cell(Atom.from_vec("H", outside_vec_8))


def test_bound_to_cell():
    a1 = Atom("N", 3, 3, 3)
    a2 = Atom("H", 1, 0, 0)
    a3 = Atom("H", 0, 1, 0)
    a4 = Atom("H", 0, 0, 1)
    mol = Molecule(a1, a2, a3, a4)
    mol.lattice_vecs = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]

    mol.wrap_to_cell()
    for atom in mol:
        assert mol.is_inside_cell(atom)
    assert mol[0] == Atom("N", 1, 1, 1)


def test_remove_duplicates():
    a1 = Atom("N", 3, 3, 3)
    a2 = Atom("H", 1, 0, 0)
    a3 = Atom("H", 0, 1, 0)
    a4 = Atom("H", 0, 0, 1)
    mol = Molecule(a1, a2, a3, a4)
    mol.lattice_vecs = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]

    mol.remove_duplicates()
    assert len(mol) == 4

    a5 = Atom("H", 1, 0, 0)
    mol.add_atom(a5)
    mol.remove_duplicates()
    assert len(mol) == 4


def test_replicated_from_xyz_str():
    a1 = Atom("O", 1, 1, 1)
    a2 = Atom("H", 1, 0, 0)
    a3 = Atom("H", 0, 1, 0)
    mol = Molecule(a1, a2, a3)
    mol.lattice_vecs = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]

    # e.g.   ‘x, y, z’, ‘-x, -y, z’, '-x, y + 1/2, -z + 1/2', ‘-2y+1/2, 3x+1/2, z-y+1/2’,
    xyz_str_1 = "x, y, z"
    mol_replicated = mol.replicated_from_xyz_str(xyz_str_1)
    assert mol == mol_replicated
    assert mol.lattice_vecs == mol_replicated.lattice_vecs

    mol = Molecule(Atom("O", 0.6, 1.1, 1.8))
    mol.lattice_vecs = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    xyz_str_2 = "y, -x, z"
    mol_replicated = mol.replicated_from_xyz_str(xyz_str_2)
    tmp_mol = Molecule(Atom("O", 1.1, 1.4, 1.8))
    assert mol_replicated == tmp_mol

    xyz_str_3 = "-x, y + 1/2, -z + 1/2"
    mol_replicated = mol.replicated_from_xyz_str(xyz_str_3)
    tmp_mol = Molecule(
        Atom("O", 1.4, 0.1, 1.2),
    )
    assert mol_replicated.is_same_geom(tmp_mol)

    xyz_str_4 = "-2y+1/2, 3x+1/2, z-y+1/2"
    mol_replicated = mol.replicated_from_xyz_str(xyz_str_4)
    tmp_mol = Molecule(Atom("O", 0.8, 0.8, 1.7))
    print(mol_replicated.atoms)
    print(tmp_mol.atoms)
    assert mol_replicated.is_same_geom(tmp_mol)



import math
from molgeom import Mat3, Vec3
from molgeom.utils import (
    lat_vecs_to_lat_params,
    lat_params_to_lat_vecs,
    cart2frac,
    frac2cart,
)


def test_lat_vecs_to_lat_params():

    a = 5.0
    b = 5.0
    c = 5.0
    alpha = 90.0
    beta = 90.0
    gamma = 90.0

    lattice_vecs = lat_params_to_lat_vecs(a, b, c, alpha, beta, gamma)

    a2, b2, c2, alpha2, beta2, gamma2 = lat_vecs_to_lat_params(lattice_vecs)

    assert math.isclose(a, a2)
    assert math.isclose(b, b2)
    assert math.isclose(c, c2)
    assert math.isclose(alpha, alpha2)
    assert math.isclose(beta, beta2)
    assert math.isclose(gamma, gamma2)

    params = (1.000000, 2.022375, 7.051241, 84.446654, 85.118719, 81.469234)
    lat_vecs = lat_params_to_lat_vecs(*params)
    tmp_mat = Mat3(
        [
            Vec3(1.000000, 0.000000, 0.000000),
            Vec3(0.300000, 2.000000, 0.000000),
            Vec3(0.600000, 0.600000, 7.000000),
        ]
    )
    for i in range(3):
        for j in range(3):
            assert math.isclose(lat_vecs[i][j], tmp_mat[i][j], rel_tol=1e-5)

    lat_params_from_tmp = lat_vecs_to_lat_params(tmp_mat)
    for i in range(len(lat_params_from_tmp)):
        assert math.isclose(params[i], lat_params_from_tmp[i], rel_tol=1e-5)


def test_conv_coords():
    lat_vecs = Mat3(
        [
            Vec3(1.000000, 0.000000, 0.000000),
            Vec3(0.300000, 2.000000, 0.000000),
            Vec3(0.600000, 0.600000, 7.000000),
        ]
    )
    cart_coords = [Vec3(0.1, 0.2, 0.3), Vec3(0.4, 0.5, 0.6), Vec3(0.7, 0.8, 0.9)]
    frac_coords = [cart2frac(x, lat_vecs) for x in cart_coords]
    cart_coords2 = [frac2cart(x, lat_vecs) for x in frac_coords]
    for i in range(3):
        for j in range(3):
            assert math.isclose(cart_coords[i][j], cart_coords2[i][j], rel_tol=1e-5)

    vec1 = Vec3(0.4814285714285711, 0.8714285714285714, 0.4285714285714286)
    vec2 = Vec3(1, 2, 3)
    vec3 = frac2cart(vec1, lat_vecs)
    vec3_matmul = lat_vecs.T() @ vec1
    assert vec3 == vec3_matmul
    print(vec3)

    vec4 = cart2frac(vec2, lat_vecs)
    vec4_matmul = lat_vecs.inv().T() @ vec2
    assert vec4 == vec4_matmul
    print(vec4)

    for i in range(3):
        assert math.isclose(vec2[i], vec3[i], rel_tol=1e-5)
        assert math.isclose(vec1[i], vec4[i], rel_tol=1e-5)



import subprocess
import platform
import tempfile
import webbrowser
import time
import shutil
from pathlib import Path
from string import Template
from importlib.resources import files


def is_jupyter_notebook():
    """
    Check if the current Python environment is running in a Jupyter Notebook.

    Returns:
        bool: True if running in Jupyter Notebook, otherwise False.
    """
    try:
        from IPython import get_ipython

        if "IPKernelApp" in get_ipython().config:
            return True
    except (ImportError, AttributeError):
        pass
    return False


def is_wsl():
    return "microsoft" in platform.uname().release.lower()


def resolve_path(file_path: Path) -> str:
    """
    If WSL, converts Linux paths to Windows-style paths.
    """

    if is_wsl():
        try:
            # use wslpath command to convert Linux path to Windows path
            result = subprocess.run(
                ["wslpath", "-w", str(file_path)],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                check=True,
            )
            win_path = result.stdout.strip()
            if win_path.startswith("\\\\"):
                win_path = win_path[2:]
            win_path = win_path.replace("\\", "/")
            return f"file://{win_path}"
        except subprocess.CalledProcessError as e:
            print(f"Error running wslpath: {e}")
            return file_path.as_uri()
    else:
        return file_path.resolve().as_uri()


def gen_mol_view_html(xyz_mol_data: str) -> str:
    """
    Generates HTML code for viewing molecular geometries using 3Dmol.js.
    args:
        xyz_mol_data: str
            XYZ-format molecular geometry data. can contain multiple molecules.
    returns:
        str: HTML string
    """

    html_template_path = files("molgeom").joinpath("templates/mol_view_template.html")
    with open(html_template_path, "r", encoding="utf-8") as f:
        html_template = Template(f.read())

    mol_js_uri = "https://cdnjs.cloudflare.com/ajax/libs/3Dmol/2.4.2/3Dmol-min.js"
    unique_id = str(time.time()).replace(".", "")
    return html_template.safe_substitute(
        {
            "mol_js_uri": mol_js_uri,
            "xyz_mol_data": xyz_mol_data,
            "unique_id": unique_id,
        }
    )


def open_html_in_browser(html_str: str, cleanup: bool = True) -> None:
    """
    Opens HTML string in a browser.
    args:
        html_str: str
            HTML string
        cleanup: bool
            If True, removes temporary file after opening (default: True)
    """

    tmp_dir = tempfile.mkdtemp()
    tmp_dir_path = Path(tmp_dir)
    # unique_id = str(time.time()).replace(".", "")
    # tmp_html_path = tmp_dir_path.joinpath(f"tmp_mol_view_{unique_id}.html")
    tmp_html_path = tmp_dir_path.joinpath("tmp_mol_view.html")
    tmp_html_path.write_text(html_str)

    if is_wsl():
        win_html_path = resolve_path(tmp_html_path)
        subprocess.run(["explorer.exe", win_html_path])
    else:
        uri = tmp_html_path.as_uri()
        webbrowser.open(uri)

    if cleanup:
        time.sleep(10)
        shutil.rmtree(tmp_dir)


def view_mol(
    xyz_mol_data: str, cleanup: bool = True, prefer_notebook: bool = True
) -> None:
    """
    View molecular geometry using 3Dmol.js in a browser.
    args:
        xyz_mol_data: str
            XYZ-format molecular geometry data. can contain multiple molecules.
        cleanup: bool
            If True, removes temporary file after opening (default: True)
        prefer_notebook: bool
            If True, tries to display in Jupyter Notebook if available (default: True)
    """

    display_in_notebook = prefer_notebook and is_jupyter_notebook()
    html_str = gen_mol_view_html(xyz_mol_data)

    if display_in_notebook:
        try:
            from IPython.display import display, HTML

            display(HTML(html_str))
        except (ImportError, ModuleNotFoundError):
            print("Failed to import and display in Jupyter Notebook.")
    else:
        open_html_in_browser(html_str, cleanup)


def test_view_mol():
    xyz_mol_data = """3
    H2O
    O  0.000000  0.000000  0.117370
    H  0.000000  0.755450 -0.469481
    H  0.000000 -0.755450 -0.469481
    5
    CH4
    C  0.000000  0.000000  0.000000
    H  0.000000 -0.000000  1.089000
    H  1.026719  0.000000 -0.363000
    H -0.513360 -0.889165 -0.363000
    H -0.513360  0.889165 -0.363000
    """
    view_mol(xyz_mol_data)


if __name__ == "__main__":
    test_view_mol()



import sys
from pathlib import Path

import click
from molgeom import Molecule, Vec3, read_file, poscar_parser

if "poscar2xyz" in sys.argv:
    if "--" not in sys.argv:
        idx = sys.argv.index("poscar2xyz") + 2
        if len(sys.argv) > idx:
            sys.argv.insert(idx, "--")


def validate_files(ctx, param, value):
    """Validate existence of input files."""
    files = []
    for file in value:
        path = Path(file.strip()).expanduser().resolve(strict=True)
        if not path.exists():
            raise click.BadParameter(f"File {file} does not exist")
        files.append(str(path))
    return files


@click.group()
def cli():
    """Molecular geometry manipulation tools."""
    pass


@cli.command()
@click.argument(
    "files", nargs=-1, type=click.Path(exists=True), required=True, callback=validate_files
)
def center(files):
    """Calculate center of mass for molecular structures."""
    for filepath in files:
        click.echo(f"\n{filepath}")
        mol = read_file(filepath)
        com = mol.center_of_mass()
        click.echo(str(com))


@cli.command()
@click.argument(
    "files", nargs=-1, type=click.Path(exists=True), required=True, callback=validate_files
)
def nuclrep(files):
    """Calculate nuclear repulsion energy."""
    for filepath in files:
        click.echo(f"\n{filepath}")
        mol = read_file(filepath)
        energy = mol.nuclear_repulsion()
        click.echo(f" {energy}")


@cli.command()
@click.argument(
    "files", nargs=-1, type=click.Path(exists=True), required=True, callback=validate_files
)
@click.option("-t", "--tol", type=float, help="Bond length tolerance (default: 0.15)")
def bonds(files, tol):
    """Analyze molecular bonds with optional tolerance."""
    tolerance = tol
    if tolerance is None:
        tolerance = click.prompt("Enter tolerance of bond length", default=0.15, type=float)

    for filepath in files:
        click.echo(f"\n{filepath}")
        mol = read_file(filepath)
        bonds = mol.get_bonds(tolerance)
        click.echo(f"Number of bonds: {len(bonds)}")
        click.echo(f"Bond tolerance: {tolerance}")
        click.echo(
            "label   bond length (Angstrom)       "
            + "atom1                                                                          -     atom2"
        )
        for label, bond_dict in enumerate(bonds):
            i, j = bond_dict["pair"]
            ai, aj = mol[(i, j)]
            dist_angst = ai.distance_to(aj)
            click.echo(
                f"{label:3d}          {dist_angst:.9f}           {i+1:3d}   {ai}    -   {j+1:3d}   {aj}"
            )


def translate_molecule(mol):
    """Translate molecule from point A to B."""
    click.echo("\n--- Translation ---")
    pA = click.prompt("Coordinate A", type=str)
    pB = click.prompt("Coordinate B", type=str)
    try:
        pA = Vec3(*map(float, pA.split()))
        pB = Vec3(*map(float, pB.split()))
        trans_vec = pB - pA
        mol.translate(trans_vec)
        return mol
    except (ValueError, TypeError) as e:
        raise click.BadParameter(f"Invalid coordinates: {e}")


def mirror_molecule(mol):
    """Mirror molecule by plane defined by three points."""
    click.echo("\n--- Reflection ---")
    try:
        p1 = Vec3(*map(float, click.prompt("Point 1", type=str).split()))
        p2 = Vec3(*map(float, click.prompt("Point 2", type=str).split()))
        p3 = Vec3(*map(float, click.prompt("Point 3", type=str).split()))
        mol.mirror_by_plane(p1, p2, p3)
        return mol
    except (ValueError, TypeError) as e:
        raise click.BadParameter(f"Invalid points: {e}")


def rotate_molecule(mol):
    """Rotate molecule around axis by angle."""
    click.echo("\n--- Rotation ---")
    try:
        p1 = Vec3(*map(float, click.prompt("Axis point 1", type=str).split()))
        p2 = Vec3(*map(float, click.prompt("Axis point 2", type=str).split()))
        angle = click.prompt("Rotation angle (degrees)", type=float)
        mol.rotate_by_axis(p1, p2, angle)
        return mol
    except (ValueError, TypeError) as e:
        raise click.BadParameter(f"Invalid input: {e}")


OP_FUNCS = {
    "translate": translate_molecule,
    "reflect": mirror_molecule,
    "rotate": rotate_molecule,
}


@cli.command()
@click.argument("file", type=click.Path(exists=True))
@click.option(
    "--operation",
    "-op",
    type=click.Choice(list(OP_FUNCS.keys())),
    help="Geometry modification operation",
)
def modify(file, operation):
    """Modify molecular geometry interactively."""
    mol = read_file(file)
    click.echo(f"\n{file}")

    if operation:
        mol = OP_FUNCS[operation](mol)
    else:
        # Get operation order
        operations = click.prompt(
            "Enter operation order (0: skip, 1-3: order) for [translate reflect rotate]", type=str
        )
        try:
            orders = [int(x) for x in operations.split()]
            if len(orders) != 3 or not all(0 <= x <= 3 for x in orders):
                raise click.BadParameter("Invalid operation order")

            op_order = dict(zip(OP_FUNCS.keys(), orders))
            for i in range(1, 4):
                for op, order in op_order.items():
                    if order == i:
                        mol = OP_FUNCS[op](mol)
        except ValueError:
            raise click.BadParameter("Invalid operation order format")

    click.echo("\nFinal molecule geometry:\n")
    click.echo(mol.to_xyz())


@cli.command()
@click.argument("file", type=click.Path(exists=True))
@click.argument("cell_range", nargs=6, type=int)
def poscar2xyz(file, cell_range):
    """
    Convert POSCAR to XYZ format with cell repetition.

    Provide 6 integers: a_min a_max b_min b_max c_min c_max

    Example:
        # 3x3x3 with original cell at center:
        molgeom poscar2xyz POSCAR -1 2 -1 2 -1 2
    """
    mol = poscar_parser(file)
    mol.replicate(
        list(cell_range[0:2]),
        list(cell_range[2:4]),
        list(cell_range[4:6]),
    )
    click.echo(len(mol))
    click.echo(f"{Path(file).stem}, rep: {cell_range}, {str(mol)}")
    click.echo(mol.to_xyz())


@cli.command()
@click.argument("file", type=click.Path(exists=True))
def split(file):
    """Split molecular clusters."""
    click.echo(f"\n{file}")
    mol = read_file(file)
    clusters = mol.get_clusters()
    click.echo(f"clusters: {clusters}")
    for cluster in clusters:
        if len(cluster) >= 3:
            click.echo(cluster)
            click.echo(cluster.to_xyz())
            click.echo()


@cli.command()
@click.argument(
    "files", nargs=-1, type=click.Path(exists=True), required=True, callback=validate_files
)
def view(files):
    """View molecular structure(s) in your browser."""
    mols = []
    for file_path in files:
        try:
            mol = Molecule.from_file(file_path)
            mols.append(mol)
        except Exception as e:
            click.echo(f"Error reading {file_path}: {str(e)}", err=True)
            continue

    if not mols:
        click.echo("No valid molecular geometry files were provided.", err=True)
        return

    Molecule.view_mols(mols)



import re
from pathlib import Path
from collections import deque

from molgeom import Vec3, Atom, Molecule, Mat3
from molgeom.parsers.parser_tools import (
    remove_trailing_empty_lines,
    validate_filepath,
    zopen,
)

from molgeom.utils.lattice_utils import lat_params_to_lat_vecs

# CIF format specification:
# http://www.physics.gov.az/book_I/S_R_Hall.pdf


def cif_tag_parser(filepath: str | Path) -> dict:
    filepath = validate_filepath(filepath)
    cif_tags = dict()
    cif_tags["filename"] = filepath.stem
    with zopen(filepath, mode="rt", encoding="utf-8") as file:
        lines = deque()
        # add empty line before loop_ block to separate tags
        for line in remove_trailing_empty_lines(file.readlines()):
            # replace tabs, non-breaking spaces, and multiple spaces with single space
            line = re.sub(r"[\s\t\xa0]+", " ", line)
            if "loop_" in line:
                lines.append(" ")
                lines.append(" ")
                lines.append(line)
            else:
                lines.append(line)
        lines.append(" ")
        lines.append(" ")

        while lines:
            line = lines.popleft()

            # load cell parameters
            if line.strip().startswith("_cell_length_a"):
                len_val_str = line.split()[1].split("(")[0]
                cif_tags["cell_length_a"] = float(len_val_str)
            if line.strip().startswith("_cell_length_b"):
                len_val_str = line.split()[1].split("(")[0]
                cif_tags["cell_length_b"] = float(len_val_str)
            if line.strip().startswith("_cell_length_c"):
                len_val_str = line.split()[1].split("(")[0]
                cif_tags["cell_length_c"] = float(len_val_str)
            if line.strip().startswith("_cell_angle_alpha"):
                angle_val_str = line.split()[1].split("(")[0]
                cif_tags["cell_angle_alpha"] = float(angle_val_str)
            if line.strip().startswith("_cell_angle_beta"):
                angle_val_str = line.split()[1].split("(")[0]
                cif_tags["cell_angle_beta"] = float(angle_val_str)
            if line.strip().startswith("_cell_angle_gamma"):
                angle_val_str = line.split()[1].split("(")[0]
                cif_tags["cell_angle_gamma"] = float(angle_val_str)

            # get tags under loop_ block
            loop_tags_idx = dict()
            if line.strip().startswith("loop_"):
                line = lines.popleft().strip()
                cnt = 0
                while line.startswith("_"):
                    tag = line.split()[0]
                    loop_tags_idx[tag] = cnt
                    line = lines.popleft().strip()
                    cnt += 1

            # load symmetry operations
            symop_tags = [
                "_space_group_symop_operation_xyz",
                "_space_group_symop.operation_xyz",
                "_symmetry_equiv_pos_as_xyz",
            ]
            symop_tags_used = [tag for tag in loop_tags_idx.keys() if tag in symop_tags]
            if symop_tags_used:
                symop_idx = loop_tags_idx[symop_tags_used[0]]
                cif_tags["symops"] = []
                while line.count(",") == 2:
                    line_str = "".join(line.split()[symop_idx:])
                    symop_str = line_str.replace("'", "")
                    cif_tags["symops"].append(symop_str)
                    line = lines.popleft().strip()

            # load atom symbols and positions
            atom_symbol_tags = [
                "_atom_site_type_symbol",
                "_atom_site_label",
            ]
            atom_fract_tags = [
                "_atom_site_fract_x",
                "_atom_site_fract_y",
                "_atom_site_fract_z",
            ]
            if (
                len(
                    {
                        atom_symbol_tag
                        for atom_symbol_tag in atom_symbol_tags
                        if atom_symbol_tag in loop_tags_idx
                    }
                )
                > 0
            ) and (
                len(
                    {
                        atom_tag
                        for atom_tag in atom_fract_tags
                        if atom_tag in loop_tags_idx
                    }
                )
                == 3
            ):
                cif_tags["atoms"] = []
                while len(line.split()) == len(loop_tags_idx):
                    splited = line.split()
                    if atom_symbol_tags[0] in loop_tags_idx:
                        symbol = splited[loop_tags_idx["_atom_site_type_symbol"]]
                        # remove charge info from symbol
                        symbol = re.sub(r"\d+[+-]?", "", symbol)
                    else:
                        symbol = splited[loop_tags_idx["_atom_site_label"]]
                        # remove label info from symbol
                        symbol = re.sub(r"\d+[+-]?", "", re.split("_", symbol)[0])
                        symbol = symbol.replace("HW", "H").replace("OW", "O")
                    fract_x = splited[loop_tags_idx["_atom_site_fract_x"]].split("(")[0]
                    fract_y = splited[loop_tags_idx["_atom_site_fract_y"]].split("(")[0]
                    fract_z = splited[loop_tags_idx["_atom_site_fract_z"]].split("(")[0]

                    atom = {
                        "symbol": symbol,
                        "fract_x": float(fract_x),
                        "fract_y": float(fract_y),
                        "fract_z": float(fract_z),
                    }
                    cif_tags["atoms"].append(atom)
                    line = lines.popleft().strip()

    if "atoms" not in cif_tags or len(cif_tags["atoms"]) == 0:
        raise ValueError("No atoms found in the CIF file")
    if any(
        tag not in cif_tags
        for tag in [
            "cell_length_a",
            "cell_length_b",
            "cell_length_c",
            "cell_angle_alpha",
            "cell_angle_beta",
            "cell_angle_gamma",
        ]
    ):
        raise ValueError("Cell parameters not found in the CIF file")
    if "symops" in cif_tags and len(cif_tags["symops"]) == 0:
        raise ValueError("No symmetry operations found in the CIF file")

    return cif_tags


def ciftag2mol(cif_tags: dict) -> Molecule:
    frac_to_cart_mat: Mat3 = lat_params_to_lat_vecs(
        cif_tags["cell_length_a"],
        cif_tags["cell_length_b"],
        cif_tags["cell_length_c"],
        cif_tags["cell_angle_alpha"],
        cif_tags["cell_angle_beta"],
        cif_tags["cell_angle_gamma"],
        angle_in_degrees=True,
    )
    mol = Molecule()
    mol.name = cif_tags["filename"]
    for atom in cif_tags["atoms"]:
        fract_vec = Vec3(atom["fract_x"], atom["fract_y"], atom["fract_z"])
        cart_vec = frac_to_cart_mat @ fract_vec
        symbol = atom["symbol"]
        atom = Atom.from_vec(symbol, cart_vec)
        mol.add_atom(atom)

    mol.lattice_vecs = frac_to_cart_mat

    return mol


def cif_parser(filepath: str | Path, apply_symop: bool = True) -> Molecule:
    filepath = validate_filepath(filepath)
    cif_tags = cif_tag_parser(filepath)
    mol = ciftag2mol(cif_tags)
    rep_mol = Molecule()
    if apply_symop and "symops" in cif_tags:
        for symop in cif_tags["symops"]:
            new_mol = mol.replicated_from_xyz_str(symop, wrap=True)
            rep_mol.merge(new_mol)
    rep_mol.lattice_vecs = mol.lattice_vecs
    return rep_mol



from .utils import FancyIndexingList, Vec3, Mat3, view_mol
from .data import consts
from .atom import Atom
from .molecule import Molecule
from .parsers import (
    cif_parser,
    gau_inp_parser,
    gms_inp_parser,
    poscar_parser,
    read_file,
    xyz_parser,
)

__all__ = [
    "Atom",
    "Molecule",
    "consts",
    "xyz_parser",
    "gau_inp_parser",
    "gms_inp_parser",
    "cif_parser",
    "poscar_parser",
    "read_file",
    "FancyIndexingList",
    "Vec3",
    "Mat3",
    "view_mol",
]



import re
from pathlib import Path
from collections import deque

from molgeom import Vec3
from molgeom.atom import Atom
from molgeom.molecule import Molecule
from molgeom.parsers.parser_tools import (
    is_valid_xyz_line,
    remove_trailing_empty_lines,
    validate_filepath,
    zopen,
)


def from_gau_inp_str(content: str) -> Molecule:
    mole = Molecule()

    # remove trailing empty lines and create a deque to pop from left
    lines = deque(remove_trailing_empty_lines(content.strip().split("\n")))

    # replace tabs, non-breaking spaces, and multiple spaces with single space
    for i in range(len(lines)):
        lines[i] = re.sub(r"[\s\t\xa0]+", " ", lines[i])

    # link0 section
    link0 = []
    while lines and lines[0].strip().startswith("%"):
        link0.append(lines.popleft().strip())

    # route section
    route_section = []
    if not lines[0].strip().startswith("#"):
        raise ValueError("Expected route section.\n" + f"Found: {lines[0]}\n")
    while lines and lines[0].strip():
        route_section.append(lines.popleft().strip())
    lines.popleft()

    # title section
    if not lines[0].strip() or lines[1].strip():
        raise ValueError(
            "Expected title section.\n"
            + f"Found: {lines[0]}\n"
            + f"Found: {lines[1]}\n"
        )
    _title = lines.popleft().strip()
    lines.popleft()

    # molecule Specification section
    if not lines[0].strip():
        raise ValueError(
            "Expected molecule specification section.\n" + f"Found: {lines[0]}\n"
        )

    try:
        charge, multiplicity = map(int, lines.popleft().strip().split())
    except ValueError as e:
        print(f"gau_parser : invalid file format \n{e}")

    # atom cartesian coordinates
    while lines and lines[0].strip():
        if not is_valid_xyz_line(lines[0]):
            raise ValueError(
                "Expected atom symbol and cartesian coordinates.\n"
                + f"Found: {lines[0]}\n"
            )
        data = lines.popleft().strip().split()
        if data[0] == "Tv":
            if not (
                lines[0].strip().startswith("Tv") and lines[1].strip().startswith("Tv")
            ):
                raise ValueError("Expected 3 lines of Tv lattice vectors.")

            lat_vec = [data[1:4]]
            for _ in range(2):
                if not is_valid_xyz_line(lines[0]):
                    raise ValueError(
                        "Expected atom symbol and cartesian coordinates.\n"
                        + f"Found: {lines[0]}\n"
                    )
                data = lines.popleft().strip().split()
                lat_vec.append(data[1:4])
            mole.lattice_vecs = [Vec3(*map(float, vec)) for vec in lat_vec]
        else:
            atom = Atom(
                symbol=data[0], x=float(data[1]), y=float(data[2]), z=float(data[3])
            )
            mole.add_atom(atom)
    if not mole:
        raise ValueError("No atoms found in file.")

    return mole


def extract_head_tail_from_gau_inp(filepath: str | Path) -> tuple[str, str]:
    file_head = []
    file_tail = []
    filepath = Path(filepath)
    if not filepath.exists():
        raise FileNotFoundError(f"{filepath} do not exist")
    if not filepath.is_file():
        raise ValueError(f"{filepath} is not a file")
    with zopen(filepath, mode="rt", encoding="utf-8") as file:
        lines = deque(file.readlines())

        # # replace tabs, non-breaking spaces, and multiple spaces with single space
        # for i in range(len(lines)):
        #     lines[i] = re.sub(r"[\s\t\xa0]+", " ", lines[i])

        # link0 section
        link0 = []
        while lines and lines[0].strip().startswith("%"):
            link0.append(lines.popleft().strip())
        file_head.extend(link0)

        # route section
        route_section = []
        if not lines[0].strip().startswith("#"):
            raise ValueError("Expected route section.\n" + f"Found: {lines[0]}\n")
        while lines and lines[0].strip():
            route_section.append(lines.popleft().strip())
        file_head.extend(route_section)
        lines.popleft()

        # title section
        if not lines[0].strip() or lines[1].strip():
            raise ValueError(
                "Expected title section.\n"
                + f"Found: {lines[0]}\n"
                + f"Found: {lines[1]}\n"
            )
        title = lines.popleft().strip()
        file_head.append(f"\n{title}\n")
        lines.popleft()

        # molecule Specification section
        if not lines[0].strip():
            raise ValueError(
                "Expected molecule specification section.\n" + f"Found: {lines[0]}\n"
            )

        try:
            charge, multiplicity = map(int, lines.popleft().strip().split())
        except ValueError as e:
            print(f"gau_parser : invalid file format \n{e}")
        file_head.append(f"{charge} {multiplicity}")

        # atom cartesian coordinates
        while lines and lines[0].strip():
            if not is_valid_xyz_line(lines[0]):
                raise ValueError(
                    "Expected atom symbol and cartesian coordinates.\n"
                    + f"Found: {lines[0]}\n"
                )
            lines.popleft().strip().split()

        # rest of the file is the tail
        for line in lines:
            file_tail.append(line.replace("\n", ""))

        return "\n".join(file_head), "\n".join(file_tail)


# Gaussian input file parser (com, gjf)
def gau_inp_parser(filepath: str | Path) -> Molecule:
    filepath = validate_filepath(filepath)
    with zopen(filepath, mode="rt", encoding="utf-8") as file:
        content = file.read()
    mol = from_gau_inp_str(content)
    mol.name = filepath.stem
    return mol



import re
import gzip
import bz2
import lzma
import warnings
from pathlib import Path


from molgeom.data.consts import ATOMIC_MASSES, SPECIAL_ELEMENTS

# atom_xyz_regex = re.compile(r"(\w\w?)(\s+[-+]?\d*\.\d+){3}")
symbol_xyz_regex = re.compile(r"(\w\w?)(\s+[-+]?(?:\d+|\d*\.\d+)){3}")
symbol_mass_xyz_regex = re.compile(r"(\w\w?)(\s+\d+\.\d+)(\s+[-+]?\d*\.\d+){3}")


def remove_trailing_empty_lines(lines: list[str]):
    while lines and not lines[-1].strip():
        lines.pop()
    return lines


def is_valid_xyz_line(line: str) -> bool:
    data = line.strip().split()
    if not data:
        return False
    if len(data) != 4:
        return False
    if not re.fullmatch(symbol_xyz_regex, line.strip()):
        return False
    if not (data[0] in ATOMIC_MASSES or data[0] in SPECIAL_ELEMENTS):
        return False
    return True


def is_valid_gms_xyz_line(line: str) -> bool:
    data = line.strip().split()
    if not data:
        return False
    if len(data) != 5:
        return False
    if not re.fullmatch(symbol_mass_xyz_regex, line.strip()):
        return False
    if data[0] not in ATOMIC_MASSES:
        return False
    return True


def validate_filepath(filepath: str | Path) -> Path:
    filepath = Path(str(filepath).strip()).expanduser().resolve(strict=True)
    if not filepath.exists():
        raise FileNotFoundError(f"{filepath} do not exist")
    if not filepath.is_file():
        raise ValueError(f"{filepath} is not a file")
    return filepath


def zopen(file: str | Path, mode=None, **kwargs):
    """
    Open a file with transparent support for compressed files (.gz, .bz2, .xz/.lzma).
    This version uses pathlib for path handling.

    Args:
        filename (str or Path): The file path.
        mode (str, optional): Mode to open the file.
            Must explicitly include 't' (text) or 'b' (binary).
            If not specified, defaults to 'r' with a warning.
        **kwargs: Additional keyword arguments for the underlying open functions.

    Returns:
        A file-like object.
    """
    filepath = Path(file)

    if mode is None:
        warnings.warn(
            "Explicit mode not specified. Defaulting to 'r'. "
            "Please specify mode including 't' (text) or 'b' (binary).",
            FutureWarning,
            stacklevel=2,
        )
        mode = "r"

    if "t" not in mode and "b" not in mode:
        warnings.warn(
            "Mode should explicitly include 't' (text) or 'b' (binary).",
            FutureWarning,
            stacklevel=2,
        )

    if "t" in mode and "encoding" not in kwargs:
        kwargs["encoding"] = "utf-8"

    ext = filepath.suffix.lower()
    if ext == ".gz":
        return gzip.open(str(filepath), mode, **kwargs)
    elif ext == ".bz2":
        return bz2.open(str(filepath), mode, **kwargs)
    elif ext in (".xz", ".lzma"):
        return lzma.open(str(filepath), mode, **kwargs)
    else:
        return filepath.open(mode, **kwargs)



from .cif import cif_parser
from .gamess import gms_inp_parser, extract_head_tail_from_gms_inp
from .gaussian import gau_inp_parser, extract_head_tail_from_gau_inp
from .parser_selector import read_file
from .poscar import poscar_parser
from .xyz import xyz_parser
from .mol import mol_parser
from .sdf import sdf_parser

__all__ = [
    "xyz_parser",
    "gau_inp_parser",
    "extract_head_tail_from_gau_inp",
    "gms_inp_parser",
    "extract_head_tail_from_gms_inp",
    "cif_parser",
    "poscar_parser",
    "mol_parser",
    "sdf_parser",
    "read_file",
]



from functools import wraps

my_Iterable = (list, tuple, set, dict)


def args_to_set(func):
    """
    Decorator to normalize the arguments of a method/function to always be a set.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], my_Iterable):
            args = set(args[0])
        elif len(args) == 1 and isinstance(args[0], str):
            args = {args[0]}
        elif len(args) > 1 and any(
            isinstance(arg, (list, tuple, set, dict)) for arg in args
        ):
            raise ValueError("Multiple arguments must be passed as a single iterable")
        else:
            args = set(args)

        return func(self, args, **kwargs)

    return wrapper


def args_to_list(func):
    """
    Decorator to normalize the arguments of a method/function to always be a list.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # # check if the first argument is an class or instance of a class
        # print(args)
        # first_arg = args[0]
        # is_class_method = inspect.isclass(first_arg)
        #
        # if is_class_method:
        #     args = args[1:]

        if len(args) == 1 and isinstance(args[0], my_Iterable):
            args = list(args[0])
        elif len(args) == 1 and isinstance(args[0], (str, bytes)):
            args = [args[0]]
        elif len(args) > 1 and any(
            isinstance(arg, (list, tuple, set, dict)) for arg in args
        ):
            raise ValueError("Multiple arguments must be passed as a single iterable")
        else:
            args = list(args)

        return func(self, args, **kwargs)

    return wrapper
